#include <stdio.h>
#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#include "define.h"
#include "RX5808.h"
#include "max7456.h"
#include "uart.h"

//******************************************************************************
//*                             declar
//******************************************************************************

RX5808 rx;
Max7456 osd;
uint8_t currentChannel = 0;
char buffer[5];                                 //Output of the itoa function

//forward declaration
uint8_t getClickType();

//******************************************************************************
//*                          AVR IO Funktion
//******************************************************************************
int freeRam () {
    extern int __heap_start, *__brkval;
    int v;
    return (int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}
uint16_t read_adc(uint8_t channel){
    ADMUX &= 0xF0;                   //Clear the older channel that was read
    ADMUX |= channel;                //Defines the new ADC channel to be read
    ADCSRA |= (1<<ADSC);             //Starts a new conversion
    while(ADCSRA & (1<<ADSC));       //Wait until the conversion is done
    return ADCW;                     //Returns the ADC value of the chosen channel
}

void printFreeMEM(){
    uart_puts_P("free: ");
    itoa(freeRam(), buffer, 10);      //Convert the read value to an ascii string
//    itoa(static_cast<uint8_t>(get_mem_unused()), buffer, 10);      //Convert the read value to an ascii string
    uart_puts(buffer);
    uart_puts_P("\n");
}

//******************************************************************************
//*                              init
//******************************************************************************

void init(){
//---------------------------------
// initialize LED pin as output
//---------------------------------
    DDRB |= (1<<LED_PIN);
    _delay_ms(20);
    
//---------------------------------
// initialize button pin as input
//---------------------------------
    DDRD &= ~(1 << BUTTON_PIN);
    PORTD |= (1<<LED_PIN);  //set internall pullup
    
//---------------------------------
// initialize analog input
//---------------------------------
    ADCSRA |= ((1<<ADPS2)|(1<<ADPS1)|(1<<ADPS0));       //16Mhz/128 = 125Khz the ADC reference clock
    ADMUX |= (1<<REFS0);                                //Voltage reference from Avcc (5v)
    ADCSRA |= (1<<ADEN);                                //Turn on ADC
    ADCSRA |= (1<<ADSC);                                //Do an initial conversion because this one is the slowest and to ensure that everything is up and running
    
//---------------------------------
// initialize timer 0
//---------------------------------
    //(8000000/8)/256 Hz = 390.000 Hz -> 2,56 ms
    //TCCR0 = (1<<CS01); // Prescaler 8
    //TIMSK |= (1<<TOIE0); // Overflow Interrupt
    sei(); // interrupt allow

//---------------------------------
// initialize uart
//---------------------------------
    uart_init( UART_BAUD_SELECT(UART_BAUD_RATE,F_CPU) );
    sei();
//---------------------------------
// OSD initialize Settings
//---------------------------------
    osd.setDisplayOffsets(45,25);   //x , y
    osd.setBlinkParams( _8fields , _3BT_BT );
    osd.activateExternalVideo();

}
//******************************************************************************
//                             RX5808
//******************************************************************************

//******************************************************************************
//* Positions in the frequency table for the 40 channels
//* Direct access via array operations does not work since data is stored in
//* flash, not in RAM. Use getPosition to retrieve data

const uint8_t positions[] PROGMEM = {
    19, 18, 32, 17, 33, 16,  7, 34,  8, 24,  6,  9, 25,  5, 35, 10, 26,  4, 11, 27,
    3, 36, 12, 28,  2, 13, 29, 37,  1, 14, 30,  0, 15, 31, 38, 20, 21, 39, 22, 23
};

uint16_t getPosition( uint8_t channel ) {
    return pgm_read_byte_near(positions + channel);
}

//******************************************************************************
//* Frequencies for the 40 channels
//* Direct access via array operations does not work since data is stored in
//* flash, not in RAM. Use getFrequency to retrieve data

const uint16_t channelFrequencies[] PROGMEM = {
    5865, 5845, 5825, 5805, 5785, 5765, 5745, 5725, // Band A - Boscam A
    5733, 5752, 5771, 5790, 5809, 5828, 5847, 5866, // Band B - Boscam B
    5705, 5685, 5665, 5645, 5885, 5905, 5925, 5945, // Band E - DJI
    5740, 5760, 5780, 5800, 5820, 5840, 5860, 5880, // Band F - FatShark \ Immersion
    5658, 5695, 5732, 5769, 5806, 5843, 5880, 5917  // Band C - Raceband
};

uint16_t getFrequency( uint8_t channel ) {
    return pgm_read_word_near(channelFrequencies + getPosition(channel));
}

//******************************************************************************
//* function: nextChannel
//******************************************************************************
uint8_t nextChannel(uint8_t channel)
{
    if (channel > (CHANNEL_MAX - 1))
        return CHANNEL_MIN;
    else
        return channel + 1;
    
}//******************************************************************************
//* function: updateScannerScreen
//*         : position = 0 to 99
//*         : value = 0 to 53
//*         : must be fast since there are frequent updates
//******************************************************************************
void updateScannerScreen(uint8_t position, uint8_t value ) {

    osd.print( read_adc(RSSI_PIN), 5, 5, 4, 0);
    
}

//******************************************************************************
//* function: drawAutoScanScreen
//******************************************************************************
void drawScanScreen( uint16_t Mhz, uint8_t channel , uint16_t rssi ) {
    
    osd.print("Auto Scanner", 7, 3);
    osd.print("Channel", 2, 6);
    osd.print(" MHz", 12, 6);
    osd.print("RSSI ", 20, 6);
    osd.print( channel, 2, 8, 2, 0);
    osd.print( Mhz, 12, 8, 4, 0);
    osd.print( rssi, 20, 8, 4, 0);

}
//******************************************************************************
//* function: drawScannerScreen
//******************************************************************************

void drawScannerScreen( void ) {
    
    osd.printMax7456Chars( bar, 25, 2, 12);
    osd.print("5.65       5.8       5.95", 2, 13);
    updateScannerScreen(0, 0);
}

//******************************************************************************
//* function: autoScan
//******************************************************************************
uint16_t autoScan( uint16_t frequency ) {
    
    uint8_t i;
    uint16_t scanRssi = 0;
    uint16_t bestRssi = 0;
    uint16_t scanFrequency;
    uint16_t bestFrequency;
    
    drawScanScreen(0,0,0);
    
    // Skip 10 MHz forward to avoid detecting the current channel
    scanFrequency = frequency + 10;
    if (!(scanFrequency % 2))
        scanFrequency++;        // RTC6715 can only generate odd frequencies
    
    // Coarse tuning
    bestFrequency = scanFrequency;
    for (i = 0; i < 60 && (scanRssi < RSSI_TRESHOLD); i++) {
        if ( scanFrequency <= (FREQUENCY_MAX - 5))
            scanFrequency += 5;
        else
            scanFrequency = FREQUENCY_MIN;
        rx.setFrequency(scanFrequency);
    _delay_ms( RSSI_STABILITY_DELAY_MS );
        scanRssi = read_adc(RSSI_PIN);
        if (bestRssi < scanRssi) {
            bestRssi = scanRssi;
            bestFrequency = scanFrequency;
        }
#ifdef debugSerialScanner
        char buffer[5];                                 //Output of the itoa function
        USART_putstring("F: ");
        itoa(bestFrequency, buffer, 10);      //Convert the read value to an ascii string
        USART_putstring(buffer);
        USART_putstring("RSSI: ");
        itoa(bestRssi, buffer, 10);                     //Convert the read value to an ascii string
        USART_putstring(buffer);                        //Send the converted value to the terminal
        USART_putstring("\r\n");

#endif
    }
    // Fine tuning
    scanFrequency = bestFrequency - 20;
    bestRssi = 0;
    for (i = 0; i < 20; i++, scanFrequency += 2) {
        rx.setFrequency(scanFrequency);
        _delay_ms( RSSI_STABILITY_DELAY_MS );
        scanRssi = read_adc(RSSI_PIN);
        if (bestRssi < scanRssi) {
            bestRssi = scanRssi;
            bestFrequency = scanFrequency;
        }
#ifdef debugSerialScanner
        USART_putstring("F: ");
        itoa(bestFrequency, buffer, 10);      //Convert the read value to an ascii string
        USART_putstring(buffer);
        USART_putstring("RSSI: ");
        itoa(bestRssi, buffer, 10);                     //Convert the read value to an ascii string
        USART_putstring(buffer);                        //Send the converted value to the terminal
        USART_putstring("\r\n");
#endif
    }

    drawScanScreen(bestFrequency,0, bestRssi);
    
    while (!getClickType()){
#ifdef debugSerialScanner
        USART_putstring("wait");
#endif
    }
    // Return the best frequency
    rx.setFrequency(bestFrequency);
    return (bestFrequency);

}

//******************************************************************************
//* function: graphicScanner
//*         : scans the 5.8 GHz band in 3 MHz increments and draws a graphical
//*         : representation. when the button is pressed the currently
//*         : scanned frequency is returned.
//******************************************************************************
uint16_t graphicScanner( uint16_t frequency ) {
    uint8_t i;
    uint16_t scanRssi;
    uint16_t bestRssi = 0;
    uint16_t scanFrequency = frequency;
    uint16_t bestFrequency = frequency;
    uint8_t rssiDisplayValue;
    
    // Draw screen frame etc
    drawScannerScreen();
    
    while ( getClickType() == NO_CLICK) {
        scanFrequency += 3;
        if (scanFrequency > FREQUENCY_MAX)
            scanFrequency = FREQUENCY_MIN;
        rx.setFrequency(scanFrequency);
        _delay_ms( RSSI_STABILITY_DELAY_MS );
        scanRssi = read_adc(1);
        rssiDisplayValue = (scanRssi - 140) / 10;    // Roughly 2 - 46
        updateScannerScreen(100 - ((FREQUENCY_MAX - scanFrequency) / 3), rssiDisplayValue );

#ifdef debugSerialScanner
        USART_putstring(": ");
        itoa(bestFrequency, buffer, 10);      //Convert the read value to an ascii string
        USART_putstring(buffer);
        USART_putstring("RSSI: ");
        itoa(bestRssi, buffer, 10);                     //Convert the read value to an ascii string
        USART_putstring(buffer);                        //Send the converted value to the terminal
        USART_putstring("\r\n");
#endif
    }
    // Fine tuning
    scanFrequency = scanFrequency - 20;
    for (i = 0; i < 20; i++, scanFrequency += 2) {
        rx.setFrequency(scanFrequency);
        _delay_ms( RSSI_STABILITY_DELAY_MS );
        scanRssi = read_adc(RSSI_PIN);
        if (bestRssi < scanRssi) {
            bestRssi = scanRssi;
            bestFrequency = scanFrequency;
        }
    }
    while (!getClickType())
    // Return the best frequency
    rx.setFrequency(bestFrequency);
    return (bestFrequency);
}
//******************************************************************************
//* function: bestChannelMatch
//*         : finds the best matching standard channel for a given frequency
//******************************************************************************
uint8_t bestChannelMatch( uint16_t frequency )
{
    int16_t comp;
    int16_t bestComp = 300;
    uint8_t bestChannel = CHANNEL_MIN;
    uint8_t i;
    
    for (i = CHANNEL_MIN; i <= CHANNEL_MAX; i++) {
        comp = abs( (int16_t)getFrequency(i) - (int16_t)frequency );
        if (comp < bestComp)
        {
            bestComp = comp;
            bestChannel = i;
        }
    }
    return bestChannel;
}

//******************************************************************************
//                              Menu
//******************************************************************************

//******************************************************************************
//* function: get_click_type
//*         : Polls the specified pin and returns the type of click that was
//*         : performed NO_CLICK, SINGLE_CLICK, DOUBLE_CLICK, LONG_CLICK
//*         : or LONG_LONG_CLICK
//******************************************************************************

uint8_t getClickType() {
    uint16_t timer = 0;
    uint8_t click_type = NO_CLICK;
    
    // check if the key has been pressed
    if (!BUTTON_PRESSED) return NO_CLICK ;
    
    while (BUTTON_PRESSED) {
        timer++;
        _delay_ms(1);
    }
    if (timer < 600)
        click_type = SINGLE_CLICK;

    if (timer >= 600)
        click_type = LONG_CLICK;
    
    // Check if there is a second click
    if ((click_type == SINGLE_CLICK) ) {
        
    }
    
    while ((!BUTTON_PRESSED) && (timer++ < 120)) {
        _delay_ms(1);
    }
    if (timer >= 200) {
        return click_type;
    }
    if (BUTTON_PRESSED ) {
        click_type = DOUBLE_CLICK;
        while (BUTTON_PRESSED) ;
    }
    return (click_type);
}

//******************************************************************************
//Flags
struct Flags {
     bool MENUEND;
     bool ACTIVEITEM;
     bool REFRESHINTERVAL;
};

struct MenuEntry ;   // dummy

typedef void (*MenuFnct) (struct MenuEntry*);
extern struct MenuEntry MainMenu[];

struct MenuEntry
{
    char*               text;
    MenuFnct            function;
    Flags               flags;
//    struct MenuEntry    *subMenu;
//    uint8_t             storage;
};

void MenuRefresh(struct MenuEntry *Menu );
MenuEntry *ptrMenu = NULL;
MenuEntry *ptrOldMenu = NULL;

void drawMenu(struct MenuEntry *Menu){
    ptrMenu = Menu;
    MenuRefresh(ptrMenu);
}

void exit(struct MenuEntry *Menu){
    osd.clearScreen();
}

void AutoSearch (struct MenuEntry *Menu)
{
    osd.clearScreen();
    currentChannel = bestChannelMatch(autoScan(getFrequency(currentChannel)));
    ptrMenu = MainMenu;
}

void BandScanner (struct MenuEntry *Menu)
{
    osd.clearScreen();
    currentChannel =  bestChannelMatch(graphicScanner(getFrequency(currentChannel)));
    ptrMenu = MainMenu;

}

void ManualMode (struct MenuEntry *Menu)
{
    uart_puts_p("ManualMode");
    while (!(getClickType() == LONG_CLICK)) {
        
        currentChannel = nextChannel( currentChannel );
        rx.setFrequency(getFrequency(currentChannel));
    }
}
/*
struct MenuEntry Settings[] =
{
    { "Batterie BEEPS",  NULL , { 0 , 0 , 0 } , NULL , 0 },
    { "Calibrate RSSI",  NULL , { 0 , 0 , 0 } , NULL , 0 },
    { "Video Input", NULL , { 0 , 0 , 0 } , NULL , 0 },
    { "Back", drawMenu , { 1 , 0 , 0} , MainMenu , 0 },
};
*/
struct MenuEntry MainMenu[] =
{
    { "Auto Scanner",  AutoSearch, { 0 , 1 , 0 } /*, NULL , 0*/},
    { "Band Scanner",  BandScanner, { 0 , 0 , 0 } /*, NULL , 0 */},
    { "Manual Mode", ManualMode, { 0 , 0 , 0 } /*, NULL , 0 */},
    { "Settings", drawMenu , { 0 , 0 , 0 } /*, NULL , 0 */},
    { "Exit", exit , { 1 , 0 , 0} /*, NULL , 0 */},

};

void MenuRefresh(struct MenuEntry *Menu )
{   //*----------------frame---------------
    osd.activateOSD();
    osd.clearScreen();
    osd.print("CYCLOPS OSD Menu", 5, 3);
    //*------------------------------------
    
    // find activ menu item
    uint8_t activItem   = 0;
    while (true) {
        if (Menu[activItem].flags.MENUEND) {
            osd.printMax7456Char(0xfC, 5, 7 + 0, true, true);
            Menu[0].flags.ACTIVEITEM = true;
            break;
        }
        activItem++;
    }
    
    // draw until the end of pointer list
    uint8_t         i = 0;
    do {
        osd.print(Menu[i].text, 7, 7+i);
        i++;
    } while (!Menu[i-1].flags.MENUEND);
    _delay_ms(20);
}

void MenuAction(struct MenuEntry *Menu ){
    uint8_t activItem   = 0;

    do{
        if (Menu[activItem].flags.ACTIVEITEM) {
//            Menu[activItem].function(Menu[activItem].subMenu);
            drawMenu(ptrMenu);
            uart_puts_p("hier kommt blödsinn raus ");
            return;
        }
        activItem++;
    }while (!Menu[activItem - 1 ].flags.MENUEND);
}

void MenuNext(struct MenuEntry *Menu ){
    uint8_t activItem   = 0;

    while (true) {
        if (Menu[activItem].flags.MENUEND && Menu[activItem].flags.ACTIVEITEM) {
            osd.printMax7456Char(0x00, 5, 7 + activItem);
            Menu[activItem].flags.ACTIVEITEM = false;
            MenuRefresh(Menu);
            return;
        }
        if (Menu[activItem].flags.ACTIVEITEM) {
            Menu[activItem].flags.ACTIVEITEM = false;
            Menu[activItem + 1].flags.ACTIVEITEM = true;
            osd.printMax7456Char(0x00, 5, 7 + activItem);
            osd.printMax7456Char(0xfC, 5, 8 + activItem, true ,true );
            return;
        }
        activItem++;
    }
}
void MenuIntervalAction(struct MenuEntry *Menu){
    
    // refresh Manager
    uint8_t activItem   = 0;
    do {
        // single refresh
        if (Menu[activItem].flags.ACTIVEITEM) {
            if (Menu[activItem].flags.REFRESHINTERVAL) {
                Menu[activItem].function(NULL);
            }
        }
    } while (!Menu[activItem - 1 ].flags.MENUEND);
}
//******************************************************************************
//*                             Main loop
//******************************************************************************

int main(void) {
    // init
    init();
    // init menu -> jump to MainMenu
    ptrMenu = MainMenu;
    MenuRefresh(ptrMenu);
    // ready
    uart_puts("Cyclops init \n\r");
    
    for(;;)
    {
//        MenuIntervalAction(ptrMenu);
        
        switch (getClickType())
        {
            case NO_CLICK: // do nothing
                break;
                
            case LONG_CLICK: //
                MenuAction(ptrMenu);
                break;
            case SINGLE_CLICK: //
                MenuNext(ptrMenu);
                break;
                
            case DOUBLE_CLICK:  //
                break;
        }

        
        
    }
    
    return 0; // never reached
}

